/**********************************************************************
 * ISO MPEG Audio Subgroup Software Simulation Group (1996)
 * ISO 13818-3 MPEG-2 Audio Encoder - Lower Sampling Frequency Extension
 *
 * $Id: encode.c,v 1.2 1998/10/05 17:06:47 larsi Exp $
 *
 * $Log: encode.c,v $
 * Revision 1.2  1998/10/05 17:06:47  larsi
 * *** empty log message ***
 *
 * Revision 1.1.1.1  1998/10/05 14:47:17  larsi
 *
 * Revision 1.1  1996/02/14 04:04:23  rowlands
 * Initial revision
 *
 * Received from Mike Coleman
 **********************************************************************/

 
#include "common.h"
#include "encoder.h"
#include "tables.h"


/************************************************************************
*
* window_subband()
*
* PURPOSE:  Overlapping window on PCM samples
*
* SEMANTICS:
* 32 16-bit pcm samples are scaled to fractional 2's complement and
* concatenated to the end of the window buffer #x#. The updated window
* buffer #x# is then windowed by the analysis window #c# to produce the
* windowed sample #z#
*
************************************************************************/

void window_subband(buffer, z, k)
short **buffer;
double z[64];
int k;
{
    typedef double XX[2][HAN_SIZE];
    static XX *x;
    double *xk;
    int i;
    static int off[2] = {0,0};
    static char init = 0;
    double t;
    double *ep0,*ep1, *ep2, *ep3, *ep4, *ep5, *ep6, *ep7;
    if (!init) {
        x = (XX *) mem_alloc(sizeof(XX),"x");
        memset(x, 0, 2*HAN_SIZE);
        init = 1;
    }
    xk=(*x)[k];

    /* replace 32 oldest samples with 32 new samples */
    for (i=0;i<32;i++) 
      xk[31-i+off[k]] = (double) *(*buffer)++/SCALE;

    ep0=&enwindow[0];
    ep1=&enwindow[64];
    ep2=&enwindow[128];
    ep3=&enwindow[192];
    ep4=&enwindow[256];
    ep5=&enwindow[320];
    ep6=&enwindow[384];
    ep7=&enwindow[448];
 
    /* shift samples into proper window positions */
    for( i = 0 ; i<64 ; i++ )
      {
        t =  xk[(i+off[k])&(512-1)] * *ep0++;
        t += xk[(i+64+off[k])&(512-1)] * *ep1++;
        t += xk[(i+128+off[k])&(512-1)] * *ep2++;
        t += xk[(i+192+off[k])&(512-1)] * *ep3++;
        t += xk[(i+256+off[k])&(512-1)] * *ep4++;
        t += xk[(i+320+off[k])&(512-1)] * *ep5++;
        t += xk[(i+384+off[k])&(512-1)] * *ep6++;
        t += xk[(i+448+off[k])&(512-1)] * *ep7++;
        z[i] = t;
      }
    
    off[k] += 480;              /*offset is modulo (HAN_SIZE-1)*/
    off[k] &= HAN_SIZE-1;

}
 
/************************************************************************
*
* filter_subband()
*
* PURPOSE:  Calculates the analysis filter bank coefficients
*
* SEMANTICS:
*      The windowed samples #z# is filtered by the digital filter matrix #m#
* to produce the subband samples #s#. This done by first selectively
* picking out values from the windowed samples, and then multiplying
* them by the filter matrix, producing 32 subband samples.
*
************************************************************************/


void
create_dct_matrix( filter )
double filter[16][32];
{
  register int i,k;
 
  for (i=0; i<16; i++)
    for (k=0; k<32; k++) {
       filter[i][k] = cos((double)((2*i+1)*k*PI64));
    }
}

void
IDCT32( xin, xout )
double *xin, *xout;
{
  int i,j;
  double s0, s1;
  typedef double MM[16][32];
  static MM *m = 0;
  if( m==0 ) {
    m = (MM *)mem_alloc(sizeof(MM),"filter");
    create_dct_matrix( *m );
  }
  for( i=0; i<16; i++ ) {
    s0 = s1 = 0.0;
    for( j=0; j<32; j+=2 ) {
      s0 += (*m)[i][j]*xin[j+0];
      s1 += (*m)[i][j+1]*xin[j+1];
    }
    xout[i] = s0+s1;
    xout[31-i] = s0-s1;
  }
}

void filter_subband(z,s)
double z[64], s[SBLIMIT];
{
   double yprime[32];
   int i;
   {
     yprime[0] = z[16];
     for( i=1; i<=16; i++ ) yprime[i] = z[i+16]+z[16-i];
     for( i=17; i<=31; i++ ) yprime[i] = z[i+16]-z[80-i];
     IDCT32( yprime, s );
   }
}

 
/************************************************************************
*
* encode_CRC
*
************************************************************************/
 
void encode_CRC(crc, bs)
unsigned int crc;
Bit_stream_struc *bs;
{
   putbits(bs, crc, 16);
}
