#include "common.h"
#include "get_audio.h"
#include "globalflags.h"
/************************************************************************
*
* get_audio()
*
* PURPOSE:  reads a frame of audio data from a file to the buffer,
*   aligns the data for future processing, and separates the
*   left and right channels
*
*
************************************************************************/
#define DWORD unsigned long 
DWORD get_audio(	                short		bufferL[1152],
					short		bufferR[1152],
					DWORD		num_samples,
					int			stereo,
					layer*		info
				)
{
	int		j;
	short	insamp[2304];
	DWORD	samples_read;
	int framesize;

	framesize = (info->version ==0) ? 576 : 1152;

	/* MPEG 1 */
	if(stereo == 2)  {
	  samples_read = read_samples(insamp, num_samples,2*framesize);
	  for(j=0;j<framesize;j++)
	    {
	      bufferL[j] = insamp[2*j];
	      bufferR[j] = insamp[2*j+1];
	    }
	}else{
	  if (autoconvert==TRUE) {
	    /* downconvert from a stereo file into a mono buffer */
	    samples_read = read_samples(insamp, num_samples,2*framesize);
	    for(j=0;j<framesize;j++)
	      {
		/* dont overflow the short int */
		bufferL[j] = ((int)insamp[2*j] + (int)insamp[2*j+1])/2;
		bufferR[j] = 0;
	      }	
	  } else {
	    samples_read = read_samples(insamp, num_samples,framesize);
	    for(j=0;j<framesize;j++)
	      {
		bufferL[j] = insamp[j];
		bufferR[j] = 0;
	      }
	  }
	}
	return(samples_read);
}











#ifdef LIBSNDFILE 
/*
** Copyright (C) 1999 Albert Faber
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
** 
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
** 
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software 
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <stdio.h>

/* External references */

static SNDFILE*	gs_pSndFileIn=NULL;
static SF_INFO	gs_wfInfo;


unsigned long GetSndSamples()
{
       return gs_wfInfo.samples;
}
int GetSndSampleRate()
{
	return gs_wfInfo.samplerate;
}

int GetSndChannels()
{
	return gs_wfInfo.channels;
}

void CloseSndFile()
{
	if (gs_pSndFileIn)
	{
		if (sf_close(gs_pSndFileIn) !=0)
		{
			fprintf(stderr, "Could not close sound file \n");
			exit(2);
		}
	}
}

void OpenSndFile(const char* lpszFileName, layer *info)
{
  if (!strcmp(lpszFileName, "-")) {
    /* Read from standard input.  assume raw PCM, default settings. */
    fprintf(stderr, "LAME compiled with LIBSNDFILE.  Compile without LIBSNDFILE to read from stdin.\n");
    exit(1);
  }  else {
    // Try to open the sound file
    gs_pSndFileIn=sf_open_read(lpszFileName,&gs_wfInfo);
  }


	// Check result
	if (gs_pSndFileIn==NULL)
	{
		fprintf(stderr, "Could not open sound file \"%s\".\n", lpszFileName);
		exit(1);
	}
#ifdef _DEBUG_SND_FILE
	printf("\n\nSF_INFO structure\n");
	printf("samplerate        :%d\n",gs_wfInfo.samplerate);
	printf("samples           :%d\n",gs_wfInfo.samples);
	printf("channels          :%d\n",gs_wfInfo.channels);
	printf("pcmbitwidth       :%d\n",gs_wfInfo.pcmbitwidth);
	printf("format            :");

	if (gs_wfInfo.format&SF_FORMAT_WAV) printf("Microsoft WAV format (big endian)");
	if (gs_wfInfo.format&SF_FORMAT_AIFF) printf("Apple/SGI AIFF format (little endian).");
	if (gs_wfInfo.format&SF_FORMAT_AU) printf("Sun/NeXT AU format (big endian).");
	if (gs_wfInfo.format&SF_FORMAT_AULE) printf("DEC AU format (little endian).");
	if (gs_wfInfo.format&SF_FORMAT_PCM) printf("PCM data in 8, 16, 24 or 32 bits.");
	if (gs_wfInfo.format&SF_FORMAT_FLOAT) printf("32 bit Intel x86 floats.");
	if (gs_wfInfo.format&SF_FORMAT_ULAW) printf("U-Law encoded.");
	if (gs_wfInfo.format&SF_FORMAT_ALAW) printf("A-Law encoded.");
	if (gs_wfInfo.format&SF_FORMAT_MS_ADPCM) printf("Microsoft ADPCM.");
	if (gs_wfInfo.format&SF_FORMAT_IMA_ADPCM) printf("IMA ADPCM.");

	printf("\n");
	printf("pcmbitwidth       :%d\n",gs_wfInfo.pcmbitwidth);
	printf("sections          :%d\n",gs_wfInfo.sections);
	printf("seekable          :\n",gs_wfInfo.seekable);
#endif
}


/************************************************************************
*
* read_samples()
*
* PURPOSE:  reads the PCM samples from a file to the buffer
*
*  SEMANTICS:
* Reads #samples_read# number of shorts from #musicin# filepointer
* into #sample_buffer[]#.  Returns the number of samples read.
*
************************************************************************/

DWORD read_samples( 
					short		sample_buffer[2304],
					DWORD		num_samples,
					DWORD		frame_size
					)
{
    DWORD			samples_read;
    DWORD			rcode;
    static DWORD	samples_to_read;

    if (frameNum==0)
	{
		samples_to_read = num_samples;
    }

    /* ignore samples_to_read for raw pcm input */
    if (ispcm || (samples_to_read >= frame_size))
	{
        samples_read = frame_size;
	}
    else
	{
        samples_read = samples_to_read;
	}
	if (gs_pSndFileIn)
		samples_read=sf_read_short(gs_pSndFileIn,sample_buffer,samples_read);
	else
		samples_read=fread(sample_buffer, sizeof(short),samples_read, stdin);


	rcode = samples_read;
	if (samples_read >=0) {
	  samples_to_read -= samples_read;
	  if (samples_read < frame_size)
	    {
	      /* printf("Insufficient PCM input for one frame - fillout with zeros\n"); */
	      for (; samples_read < frame_size; sample_buffer[samples_read++] = 0);
	      samples_to_read = 0;
	    }
	}
	return(rcode);
}


#else /* ifdef LIBSNDFILE */

/************************************************************************
 ************************************************************************
 ************************************************************************
 ************************************************************************
 ************************************************************************
 ************************************************************************
 *
 * OLD ISO/LAME routines follow.  Used if you dont have LIBSNDFILE
 *
 ************************************************************************
 ************************************************************************
 ************************************************************************
 ************************************************************************
 ************************************************************************
 ************************************************************************/

static FILE *musicin=NULL;
static unsigned long num_samples;
static int samp_freq;
static int num_channels;

void CloseSndFile()
{
  if (fclose(musicin) != 0){
    fprintf(stderr, "Could not close audio input file\n");
    exit(2);
  }
}


unsigned long GetSndSamples()
{
       return num_samples;
}
int GetSndSampleRate()
{
	return samp_freq;
}

int GetSndChannels()
{
	return num_channels;
}

void OpenSndFile(const char* inPath, layer *info)
{
  int isaiff;

  /* set the defaults from info incase we cannot determine them from file */
  num_samples=MAX_U_32_NUM;
  samp_freq=1000*s_freq[info->version][info->sampling_frequency];
  num_channels = (info->mode == MPG_MD_MONO) ? 1 : 2;

  if (!strcmp(inPath, "-")) {
    /* Read from standard input. */
    musicin = stdin;
  } else {
    if ((musicin = fopen(inPath, "rb")) == NULL) {
      fprintf(stderr, "Could not find \"%s\".\n", inPath);
      exit(1);
    }
  }

  /* check for AIFF and parse AIFF headers */
  if (!ispcm) {
    isaiff = parse_aiff(inPath,info);
    if (!isaiff) {
      /* check for WAVE and parse headers */
      iswav = parse_wavheader();
    }
    if (!(isaiff || iswav)) ispcm=1;
  }
  
  if (ispcm) {
    /* assume raw PCM */
    fprintf(stderr, "Assuming raw pcm input file");
    if (swapbytes==TRUE)
      fprintf(stderr, " : Forcing byte-swapping\n");
    else
      fprintf(stderr, "\n");
  }
}


/************************************************************************
*
* read_samples()
*
* PURPOSE:  reads the PCM samples from a file to the buffer
*
*  SEMANTICS:
* Reads #samples_read# number of shorts from #musicin# filepointer
* into #sample_buffer[]#.  Returns the number of samples read.
*
************************************************************************/

unsigned long read_samples(sample_buffer, num_samples, frame_size)
short sample_buffer[2304];
unsigned long num_samples, frame_size;
{
    unsigned long samples_read;
    unsigned long rcode;
    static unsigned long samples_to_read;
    extern int swapbytes; /* if TRUE then force swapping of byte order */

    if (frameNum==0) {
        samples_to_read = num_samples;
    }
    /* ignore samples_to_read for raw pcm input */
    if (ispcm || (samples_to_read >= frame_size))
        samples_read = frame_size;
    else
        samples_read = samples_to_read;

    if ((samples_read =
         fread(sample_buffer, sizeof(short), (int)samples_read, musicin)) == 0)
      { /* printf("Hit end of audio data\n"); */  }

    /*
       Samples are big-endian. If this is a little-endian machine
       we must swap
     */
    if ( NativeByteOrder == order_unknown )
      {
	NativeByteOrder = DetermineByteOrder();
	if ( NativeByteOrder == order_unknown )
	  {
	    fprintf( stderr, "byte order not determined\n" );
	    exit( 1 );
	  }
      }
    if (!iswav && ( NativeByteOrder == order_littleEndian ))
      SwapBytesInWords( sample_buffer, samples_read );

    if (iswav && ( NativeByteOrder == order_bigEndian ))
      SwapBytesInWords( sample_buffer, samples_read );

    if (swapbytes==TRUE)
      SwapBytesInWords( sample_buffer, samples_read );

    /* dont count samples for pcm files */
    if (!ispcm) samples_to_read -= samples_read;

    rcode=samples_read;
    if (samples_read < frame_size && samples_read >= 0) {
      /* printf("Insufficient PCM input for one frame - fillout with zeros\n"); */
        for (; samples_read < frame_size; sample_buffer[samples_read++] = 0);
        samples_to_read = 0;
    }
    return(rcode);
}


/*****************************************************************************
 *
 *  Read Audio Interchange File Format (AIFF) headers.
 *
 *****************************************************************************/

int aiff_read_headers( FILE *file_ptr, IFF_AIFF *aiff_ptr )
{
    int chunkSize, subSize, sound_position;
    
    if ( fseek(file_ptr, 0, SEEK_SET) != 0 )
        return -1;
    
    if ( Read32BitsHighLow(file_ptr) != IFF_ID_FORM )
        return -1;
    
    chunkSize = Read32BitsHighLow( file_ptr );
    
    if ( Read32BitsHighLow(file_ptr) != IFF_ID_AIFF )
        return -1;
    
    sound_position = 0;
    while ( chunkSize > 0 )
    {
        chunkSize -= 4;
        switch ( Read32BitsHighLow(file_ptr) )
	{
            
          case IFF_ID_COMM:
            chunkSize -= subSize = Read32BitsHighLow( file_ptr );
            aiff_ptr->numChannels = Read16BitsHighLow( file_ptr );
            subSize -= 2;
            aiff_ptr->numSampleFrames = Read32BitsHighLow( file_ptr );
            subSize -= 4;
            aiff_ptr->sampleSize = Read16BitsHighLow( file_ptr );
            subSize -= 2;
            aiff_ptr->sampleRate  = ReadIeeeExtendedHighLow( file_ptr );
            subSize -= 10;
            while ( subSize > 0 )
	    {
                getc( file_ptr );
                subSize -= 1;
	    }
            break;
            
          case IFF_ID_SSND:
            chunkSize -= subSize = Read32BitsHighLow( file_ptr );
            aiff_ptr->blkAlgn.offset = Read32BitsHighLow( file_ptr );
            subSize -= 4;
            aiff_ptr->blkAlgn.blockSize = Read32BitsHighLow( file_ptr );
            subSize -= 4;
            sound_position = ftell( file_ptr ) + aiff_ptr->blkAlgn.offset;
            if ( fseek(file_ptr, (long) subSize, SEEK_CUR) != 0 )
                return -1;
            aiff_ptr->sampleType = IFF_ID_SSND;
            break;
            
          default:
            chunkSize -= subSize = Read32BitsHighLow( file_ptr );
            while ( subSize > 0 )
	    {
                getc( file_ptr );
                subSize -= 1;
	    }
            break;
	}
    }
    return sound_position;
}

#include "wavheader.h"
#define ENDSWAP_SHORT(x) ((((x)>>8)&0xFF)|(((x)&0xFF)<<8))
#define ENDSWAP_INT(x) ((((x)>>24)&0xFF)|(((x)>>8)&0xFF00)|(((x)&0xFF00)<<8)|(((x)&0xFF)<<24))
void endswap_wave_fmt (WaveHeader *wavfmt) {
	wavfmt->main_chunk	= ENDSWAP_INT	(wavfmt->main_chunk);
	wavfmt->length		= ENDSWAP_INT	(wavfmt->length);
	wavfmt->chunk_type	= ENDSWAP_INT	(wavfmt->chunk_type);

	wavfmt->sub_chunk	= ENDSWAP_INT	(wavfmt->sub_chunk);
	wavfmt->sc_len		= ENDSWAP_INT	(wavfmt->sc_len);
	wavfmt->format		= ENDSWAP_SHORT	(wavfmt->format);
	wavfmt->modus		= ENDSWAP_SHORT	(wavfmt->modus);
	wavfmt->sample_fq	= ENDSWAP_INT	(wavfmt->sample_fq);
	wavfmt->byte_p_sec	= ENDSWAP_INT	(wavfmt->byte_p_sec);
	wavfmt->byte_p_spl	= ENDSWAP_SHORT	(wavfmt->byte_p_spl);
	wavfmt->bit_p_spl	= ENDSWAP_SHORT	(wavfmt->bit_p_spl);

	//	wavfmt->data_chunk	= ENDSWAP_INT	(wavfmt->data_chunk);
	//	wavfmt->data_length	= ENDSWAP_INT	(wavfmt->data_length);
}

int parse_aiff(const char inPath[],layer *info)
{
  IFF_AIFF pcm_aiff_data;
  long soundPosition;
  int isaiff;

  isaiff = ((soundPosition = aiff_read_headers(musicin, &pcm_aiff_data)) != -1) ;
  if (isaiff) {
    fprintf(stderr, "Found Audio IFF sound file headers: ");
    aiff_check(inPath, &pcm_aiff_data, &info->version);
    fseek(musicin, soundPosition, SEEK_SET);  /* ignore errors */
    samp_freq =  pcm_aiff_data.sampleRate;
    fprintf(stderr, "%.1f kHz samples\n",
	    pcm_aiff_data.sampleRate/1000);
    /* Determine number of samples in sound file */
    num_samples = pcm_aiff_data.numSampleFrames;
    num_channels = pcm_aiff_data.numChannels;
  }else{
    fseek(musicin, 0, SEEK_SET);  /* rewind to beginning, ignore errors */
  }

  return isaiff;
}

int parse_wavheader()
{
  int iswav=0;
  WaveHeader wavhead;

  fread(&wavhead,sizeof(WaveHeader)-8,1,musicin);
  if (DetermineByteOrder()==order_bigEndian) 
    endswap_wave_fmt(&wavhead);
  iswav = (wavhead.chunk_type==WAVE);


  /* read and ignore the rest of the chunks until 'data' */
  do {
    if (!fread(&wavhead.data_chunk,4,1,musicin)) break;
    if (!fread(&wavhead.data_length,4,1,musicin)) break;
    if (DetermineByteOrder()==order_bigEndian) {
      wavhead.data_chunk= ENDSWAP_INT	(wavhead.data_chunk);
      wavhead.data_length= ENDSWAP_INT	(wavhead.data_length);
    }
    if (wavhead.data_chunk==DATA) break;
    fseek(musicin, 4, SEEK_CUR);
  }while (1);


  if (iswav) {

    fprintf(stderr,"Found WAV header:  %3.1fkHz %s\n",
	    wavhead.sample_fq/1000.0,
	    (wavhead.modus==WAVE_MONO ? "mono" : "stereo"));
    /*    
    fprintf(stderr," Format: %s\n",wavhead.format==PCM_CODE ? "PCM" : "unknown");
    fprintf(stderr," Mode  : %s\n",wavhead.modus==WAVE_STEREO ? "stereo" : "mono");
    fprintf(stderr," Freq  : %d\n",wavhead.sample_fq);
    fprintf(stderr," Bits  : %d\n",wavhead.bit_p_spl);
    fprintf(stderr," FLen  : %d\n",wavhead.length);
    fprintf(stderr," data  : %d\n",wavhead.data_length);
    */

    if (wavhead.bit_p_spl != 16) {
      fprintf(stderr," Bits  : %d not supported!\n",wavhead.bit_p_spl);
      exit(1);
    }
    num_channels=2;
    if (wavhead.modus==WAVE_MONO) num_channels=1;
    samp_freq = wavhead.sample_fq;
    num_samples = (wavhead.data_length / (num_channels * wavhead.bit_p_spl / 8)) ;
  }else{
    fseek(musicin, 0, SEEK_SET);  /* rewind to beginning, ignore errors */
  }

  
  return iswav;
}
/************************************************************************
*
* aiff_check
*
* PURPOSE:  Checks AIFF header information to make sure it is valid.
*           Exits if not.
*
************************************************************************/

void aiff_check(const char *file_name, IFF_AIFF *pcm_aiff_data, int *version)
{
    if (pcm_aiff_data->sampleType != IFF_ID_SSND) {
       fprintf(stderr, "Sound data is not PCM in \"%s\".\n", file_name);
       exit(1);
    }

    if(SmpFrqIndex((long)pcm_aiff_data->sampleRate, version) < 0) {
       fprintf(stderr, "in \"%s\".\n", file_name);
       exit(1);
    }

    if (pcm_aiff_data->sampleSize != sizeof(short) * BITS_IN_A_BYTE) {
        fprintf(stderr, "Sound data is not %d bits in \"%s\".\n",
		(unsigned int) sizeof(short) * BITS_IN_A_BYTE, file_name);
        exit(1);
    }

    if (pcm_aiff_data->numChannels != MONO &&
        pcm_aiff_data->numChannels != STEREO) {
       fprintf(stderr, "Sound data is not mono or stereo in \"%s\".\n",
	       file_name);
       exit(1);
    }

    if (pcm_aiff_data->blkAlgn.blockSize != 0) {
       fprintf(stderr, "Block size is not %d bytes in \"%s\".\n",
	       0, file_name);
       exit(1);
    }

    if (pcm_aiff_data->blkAlgn.offset != 0) {
       fprintf(stderr, "Block offset is not %d bytes in \"%s\".\n",
	       0, file_name);
       exit(1);
    }
}


#endif


