#ifndef RH

/************************************************************************
 * allocate bits among the 4 granules based on PE
 * mt 6/99
 ************************************************************************/
int VBR_on_pe(frame_params *fr_ps, layer *info, III_side_info_t *l3_side, 
	       int VBRbits[2][2],double pe[2][2], int *mean_bits)

{ 
  int nshort=0;
  int index,gr,ch;
  int bitsPerFrame;
  int stereo = fr_ps->stereo;  
  int fullframebits;
  int mode_gr = (info->version == 1) ? 2 : 1;

  getframebits(info,stereo,&bitsPerFrame,mean_bits);
  fullframebits= ResvFrameBegin( fr_ps, l3_side, *mean_bits, bitsPerFrame );
  
  
  for (gr=0;gr<mode_gr;gr++)
    for (ch=0;ch<stereo;ch++) {
      int shortblock = (l3_side->gr[gr].ch[ch].tt.block_type==2);
      if (shortblock) nshort ++;
    }
  index = nshort;
  if (index>3) index=3; 

  /* impose a minimum bit rate based on number of short blocks */  
  if (index>0) {
    /* nshort=1: min 160kbs.  (1 shortblock granule)
     * nshort=2: min 190kbs.  (2 shortblock granules)
     * nshort=3: min 220kbs   (3 shortblock granules)
     * nshort=4: min 256kbs   (4 shortblock granules) */
    if (info->bitrate_index < 9 + index) info->bitrate_index = 9+index;  
    if (info->bitrate_index > 14) info->bitrate_index=14;

    
    getframebits(info,stereo,&bitsPerFrame,mean_bits);
    fullframebits= ResvFrameBegin( fr_ps, l3_side, *mean_bits, bitsPerFrame );
  }

  /* allocate a minimum 125 bits per channel (approx 32kbs) */
  for (gr=0;gr<mode_gr;gr++) {
    for (ch=0;ch<stereo;ch++) {
      VBRbits[gr][ch] = 125;
    }
  }
  fullframebits -= stereo * mode_gr * 125;

  /* divide bits based on PE */  
  if (fullframebits > 0 )
  {
    double pe2[2][2],pe_tot=.01;
    for (gr=0;gr<mode_gr;gr++) {
      for (ch=0;ch<stereo;ch++) {
	int shortblock = (l3_side->gr[gr].ch[ch].tt.block_type==2);
	pe2[gr][ch] = Min(pe[gr][ch],2800.);
	if (shortblock)  pe2[gr][ch] = Max(pe2[gr][ch],1800.0);
	pe_tot += pe2[gr][ch];
      }
    }
    for (gr=0;gr<mode_gr;gr++) {
      for (ch=0;ch<stereo;ch++) {
	VBRbits[gr][ch] = (fullframebits*pe2[gr][ch])/pe_tot;
      }
    }
  }
  /* return value = 1 => do not decrease bit rate further */
  if (index) 
    return (info->bitrate_index <= 9 + index);
  else return 0;
}




/************************************************************************
 * increase or decrease frame bitrate based on max_over = number of scale
 * factor bands with audible distortion.  If bitrate in increased, allocate
 * the additional bits among the granules based on "over"
 * mt 6/99
 ************************************************************************/
void VBR_on_over(frame_params *fr_ps, layer *info, III_side_info_t *l3_side, 
	       int VBRbits[2][2], double pe[2][2], int max_over, int VBR_q,
	       int over[2][2], int *mean_bits, int *VBR_no_decrease)

{ 
  int extrabits=0;
  int gr,ch;
  int bitsPerFrame;
  int stereo = fr_ps->stereo;  
  int fullframebits;
  int mode_gr = (info->version == 1) ? 2 : 1;

  if (max_over <= VBR_q) {
    info->bitrate_index --;
    if (info->bitrate_index < VBR_min_bitrate) 
      info->bitrate_index=VBR_min_bitrate;
    *VBR_no_decrease = VBR_on_pe(fr_ps,info,l3_side,VBRbits,pe,mean_bits);
    return;
  }

  getframebits(info,stereo,&bitsPerFrame,mean_bits);
  fullframebits= ResvFrameBegin( fr_ps, l3_side, *mean_bits, bitsPerFrame );

  info->bitrate_index ++;
  if (info->bitrate_index > VBR_max_bitrate) 
    info->bitrate_index=VBR_max_bitrate;
  
  getframebits(info,stereo,&bitsPerFrame,mean_bits);
  extrabits= ResvFrameBegin( fr_ps, l3_side, *mean_bits, bitsPerFrame );
  extrabits -= fullframebits;

  /* divide extra bits based on OVER */
  if (extrabits > 0) {
    double over2[2][2],over_tot=0;
    for (gr=0;gr<mode_gr;gr++) {
      for (ch=0;ch<stereo;ch++) {
	over2[gr][ch] = over[gr][ch] + 1;
	over_tot += over2[gr][ch];
      }
    }
    for (gr=0;gr<mode_gr;gr++) {
      for (ch=0;ch<stereo;ch++) {
	VBRbits[gr][ch] += (extrabits*over2[gr][ch])/over_tot;
      }
    }
  }
  for (gr=0;gr<mode_gr;gr++) {
    for (ch=0;ch<stereo;ch++) {
      if (VBRbits[gr][ch]>4095) VBRbits[gr][ch]=4095;
    }
  }
}













/************************************************************************/
/*  iteration_loop()                                                    */
/************************************************************************/
void
VBR_iteration_loop( double pe[2][2], double ms_ener_ratio[2],
		double xr_org[2][2][576], III_psy_ratio *ratio,
		III_side_info_t *l3_side, int l3_enc[2][2][576],
		III_scalefac_t *scalefac, frame_params *fr_ps)
{
  III_psy_xmin l3_xmin;
  layer *info;
  int VBRbits[2][2];
  int best_over[2][2];
  double noise[4];
  double targ_noise[4];
  int mean_bits;
  int stereo = fr_ps->stereo;
  int ch, gr, i, mode_gr;
  int VBR_no_decrease;
  double xr[2][2][576];


  iteration_init(xr_org,l3_side,l3_enc,fr_ps,&l3_xmin);
  info = fr_ps->header;
  mode_gr = (info->version == 1) ? 2 : 1;

  VBR_no_decrease=0;
  VBR_no_decrease=VBR_on_pe(fr_ps,info,l3_side,VBRbits,pe,&mean_bits);


  /* quantize! */
  do {
    int max_over, VBR_q2;
    VBR_q2=VBR_q;
    /* Robert Hegemann 7/99 
Distortions in frames at lower bitrates are more anoying than at higher bitrates. 
The user selects the allowed number of distortions at 112 kbs
frames and defines therefore the maximum for all other possible
bitrates. 
    */
    VBR_q2 = ((int)(0.5+
    (double)(bitrate[info->version][info->lay-1][info->bitrate_index])
  / (double)(bitrate[info->version][info->lay-1][8])
  * (double)(VBR_q)   ));


    for ( gr = 0; gr < mode_gr; gr++ ) {
      int targ_bits[2];
      /* copy data to be quantized into xr */
      if (convert_mdct) ms_convert(xr[gr],xr_org[gr]);
      else memcpy(xr[gr],xr_org[gr],sizeof(double)*2*576);   
      
      for (ch=0 ; ch < stereo ; ch ++) targ_bits[ch]=VBRbits[gr][ch];
      if (reduce_sidechannel) 
	reduce_side(targ_bits,ms_ener_ratio[gr],mean_bits);
      
      for (ch=0 ; ch < stereo ; ch ++) {
	outer_loop( xr, targ_bits[ch], noise, targ_noise, 0,
	    &l3_xmin,l3_enc, fr_ps, 
	    scalefac,gr,stereo, l3_side, ratio, ms_ener_ratio[gr], ch);
	best_over[gr][ch]=noise[0];
      }
    }
      

    /* see if we should try a higher bitrate quantization */
    max_over=0;
    if (reduce_sidechannel) {
      /* for this case, allow a little extra distortion in the side channel */
      for ( gr = 0; gr < mode_gr; gr++ ) {
	double fac, m;
	fac = 2*(.5-ms_ener_ratio[gr]);
	m   = nint(best_over[gr][1] - fac);
	max_over = Max( max_over, best_over[gr][0] );
	max_over = Max( max_over, m );
      }
    } else {
      for ( gr = 0; gr < mode_gr; gr++ ) {
	for (ch=0 ; ch < stereo ; ch ++ ) {
 	  if (best_over[gr][ch]>max_over) max_over=best_over[gr][ch];
	}
      }
    }
    
    
    if ((max_over <= VBR_q2) && (info->bitrate_index <= VBR_min_bitrate)) break;
    if ((max_over > VBR_q2) && (info->bitrate_index >= VBR_max_bitrate)) break;
    /* do not decrease bit rate if we have ever increased bit rate 
     * otherwise infinite loop */
    if ((max_over <= VBR_q2) && (VBR_no_decrease)) break;
    
    /* compute a new bitrate, and allocate bits between granules */
    if (max_over > VBR_q2) VBR_no_decrease=1;
    VBR_on_over(fr_ps,info,l3_side,VBRbits,pe,max_over,VBR_q2,best_over,&mean_bits,&VBR_no_decrease);
  } while (1);
  

  /* update reservoir status after FINAL quantization/bitrate */
  for ( gr = 0; gr < mode_gr; gr++ ) {
    for (ch=0 ; ch < stereo ; ch ++ ) {
      gr_info *cod_info = &l3_side->gr[gr].ch[ch].tt;
      ResvAdjust( fr_ps, cod_info, l3_side, mean_bits );
    }
  }

  
  /* set the sign of l3_enc */
  for ( gr = 0; gr < mode_gr; gr++ )
    for ( ch =  0; ch < stereo; ch++ )
      {
	int *pi = &l3_enc[gr][ch][0];
	for ( i = 0; i < 576; i++) 	    {
	  double pr = xr[gr][ch][i];
	  if ( (pr < 0) && (pi[i] > 0) )   pi[i] *= -1;
	}
      }

  ResvFrameEnd( fr_ps, l3_side, mean_bits );
}
#endif






/************************************************************************/
/*  init_outer_loop  mt 6/99                                            */
/************************************************************************/
void init_outer_loop_dual(
    double xr[2][2][576],        /*  could be L/R OR MID/SIDE */
    double xr_org[2][2][576],
    III_psy_xmin  *l3_xmin,   /* the allowed distortion of the scalefactor */
    III_scalefac_t *scalefac, /* scalefactors */
    int gr, int stereo, III_side_info_t *l3_side,
    III_psy_ratio *ratio, int ch)
{
  int sfb,i;
  gr_info *cod_info;  
  cod_info = &l3_side->gr[gr].ch[ch].tt;

  /* compute max allowed distortion */
  if (convert_psy) /* l/r thresholds, use unconverted data */
    calc_xmin( xr_org, ratio, cod_info, l3_xmin, gr, ch );
  else
    calc_xmin( xr, ratio, cod_info, l3_xmin, gr, ch );
  
  /* if ( info->version == 1 )
     calc_scfsi( xr[gr][ch], l3_side, &l3_xmin, ch, gr ); 
  */
  
    
  /* reset of iteration variables */
    
  for ( sfb = 0; sfb < SFB_LMAX; sfb++ )
    scalefac->l[gr][ch][sfb] = 0;
  for ( sfb = 0; sfb < SFB_SMAX; sfb++ )
    for ( i = 0; i < 3; i++ )
      scalefac->s[gr][ch][sfb][i] = 0;
  
  for ( i = 0; i < 4; i++ )
    cod_info->slen[i] = 0;
  cod_info->sfb_partition_table = &nr_of_sfb_block[0][0][0];
  
  cod_info->part2_3_length    = 0;
  cod_info->big_values        = 0;
  cod_info->count1            = 0;
  cod_info->scalefac_compress = 0;
  cod_info->table_select[0]   = 0;
  cod_info->table_select[1]   = 0;
  cod_info->table_select[2]   = 0;
  cod_info->subblock_gain[0]  = 0;
  cod_info->subblock_gain[1]  = 0;
  cod_info->subblock_gain[2]  = 0;
  cod_info->region0_count     = 0;
  cod_info->region1_count     = 0;
  cod_info->part2_length      = 0;
  cod_info->preflag           = 0;
  cod_info->scalefac_scale    = 0;
  cod_info->quantizerStepSize = 0.0;
  cod_info->count1table_select= 0;
  cod_info->address1          = 0;
  cod_info->address2          = 0;
  cod_info->address3          = 0;
  
  
  if (experimentalZ) {
    /* compute subblock gains */
    int j,b;  double en[3],mx;
    if ((cod_info->block_type ==2) ) {
      /* estimate energy within each subblock */
      for (b=0; b<3; b++) en[b]=0;
      for ( i=0,j = 0; j < 192; j++ ) {
	for (b=0; b<3; b++) {
	  en[b]+=xr[gr][ch][i]*xr[gr][ch][i];
	  i++;
	}
      }
      mx = 1e-12;
      for (b=0; b<3; b++) mx=Max(mx,en[b]);
      for (b=0; b<3; b++) en[b] = Max(en[b],1e-12)/mx;
      //	printf("ener = %4.2f  %4.2f  %4.2f  \n",en[0],en[1],en[2]);
      /* pick gain so that 2^(2gain)*en[0] = 1  */
      /* gain = .5* log( 1/en[0] )/log(2) = -.5*log(en[])/log(2) */
      for (b=0; b<3; b++) {
	cod_info->subblock_gain[b]=nint(-.5*log(en[b])/log(2.0));
	if (cod_info->subblock_gain[b] > 2) 
	  cod_info->subblock_gain[b]=2;
	if (cod_info->subblock_gain[b] < 0) 
	  cod_info->subblock_gain[b]=0;
      }
    }
  }
}




void quant_compare_dual(int better[2], int notdone[2], int stereo, 
double ms_ener_ratio,
int best_over[2],double best_tot_noise[2],double best_over_noise[2],double best_max_noise[2],
int over[2],double tot_noise[2], double over_noise[2], double max_noise[2]) 
{
  int ch;
  /*
    noise is given in decibals (db) relative to masking thesholds.

    over_noise:  sum of quantization noise > masking
    tot_noise:   sum of all quantization noise
    max_noise:   max quantization noise 

   */


  for (ch=0 ; ch < stereo ; ch ++ ) {
    better[ch]=0;
    if (notdone[ch]) {
      if (convert_psy) {
	better[ch] = (over[0]+over[1]) < (best_over[0]+best_over[1]);
	if ((over[0]+over[1])==(best_over[0]+best_over[1])) {
	  better[ch] = (over_noise[0]+over_noise[1]) < (best_over_noise[0]+best_over_noise[1]);
	}
      }else{
	better[ch] = ((over[ch] < best_over[ch]) ||
		      ((over[ch]==best_over[ch]) && (over_noise[ch]<best_over_noise[ch])) ) ;
      }
    }
  }
}
  



/************************************************************************/
/*  outer_loop                                                         */
/************************************************************************/
/*  Function: The outer iteration loop controls the masking conditions  */
/*  of all scalefactorbands. It computes the best scalefac and          */
/*  global gain. This module calls the inner iteration loop             
 * 
 *  mt 5/99 completely rewritten to allow for bit reservoir control,   
 *  mid/side channels with L/R or mid/side masking thresholds, 
 *  and chooses best quantization instead of last quantization when 
 *  no distortion free quantization can be found.  
 *  
 *  added VBR support mt 5/99
 ************************************************************************/
void outer_loop_dual(
    double xr[2][2][576],        /*  could be L/R OR MID/SIDE */
    double xr_org[2][2][576],
    int mean_bits,
    int VBRbits[2][2],
    int bit_rate,
    int best_over[2],
    III_psy_xmin  *l3_xmin,   /* the allowed distortion of the scalefactor */
    int l3_enc[2][2][576],    /* vector of quantized values ix(0..575) */
    frame_params *fr_ps,
    III_scalefac_t *scalefac, /* scalefactors */
    int gr, int stereo, III_side_info_t *l3_side,
    III_psy_ratio *ratio, double pe[2][2], double ms_ener_ratio[2])
{
  int status[2],notdone[2]={0,0},count[2]={0,0},bits_found[2];
  int targ_bits[2],real_bits[2],tbits,extra_bits; 
  int scalesave_l[2][CBLIMIT], scalesave_s[2][CBLIMIT][3];
  int sfb, bits, huff_bits, save_preflag[2], save_compress[2];
  double xfsf[2][4][CBLIMIT];
  double xrpow[2][2][576],temp;
  double distort[2][4][CBLIMIT];
  int save_l3_enc[2][576];  
  int save_real_bits[2];
  int i,over[2], iteration, ch, compute_stepsize;
  int better[2];
  double max_noise[2];
  double best_max_noise[2];
  double over_noise[2];
  double best_over_noise[2];
  double tot_noise[2];
  double best_tot_noise[2];
  gr_info save_cod_info[2];
  gr_info *cod_info[2];  

  cod_info[0] = &l3_side->gr[gr].ch[0].tt;
  cod_info[1] = &l3_side->gr[gr].ch[1].tt;
  for (ch=0 ; ch < stereo ; ch ++) 
    best_over[ch] = 100;
      

  if (convert_mdct) ms_convert(xr[gr],xr_org[gr]);
  else memcpy(xr[gr],xr_org[gr],sizeof(double)*2*576);   
  for (ch=0; ch<stereo; ch++)
    init_outer_loop_dual(xr,xr_org,l3_xmin,scalefac,gr,stereo,l3_side,ratio,ch);  
  
  for (ch=0 ; ch < stereo ; ch ++) {
    count[ch]=0;
    for (i=0; i<576; i++) {
      if ( fabs(xr[gr][ch][i]) > 0 ) count[ch]++; 
    }
    notdone[ch]=count[ch];
    if (count[ch]==0) best_over[ch]=0;
  }

  /******************************************************************
   * allocate bits for each channel 
   ******************************************************************/
  if (VBR) {
    for (ch=0 ; ch < stereo ; ch ++ )
      targ_bits[ch]=VBRbits[gr][ch];

  }else { 
    int add_bits[2]; 

    /* allocate targ_bits for granule */
    ResvMaxBits2( mean_bits, &tbits, &extra_bits, gr);

    for (ch=0 ; ch < stereo ; ch ++ )
      targ_bits[ch]=tbits/stereo;

    
    // allocate extra bits from reservoir based on PE 
    bits=0;
    for (ch=0; ch<stereo; ch++) {
      double pe_temp=pe[gr][ch];
      if (convert_psy) pe_temp=Max(pe[gr][0],pe[gr][1]);

      /* extra bits based on PE > 700 */
      add_bits[ch]=(pe_temp-750)/1.4;  /* 2.0; */

      
      /* short blocks need extra, no matter what the pe */
      if (cod_info[ch]->block_type==2) 
	if (add_bits[ch]<500) add_bits[ch]=500;
      
      if (add_bits[ch] < 0) add_bits[ch]=0;
      bits += add_bits[ch];
    }
    for (ch=0; ch<stereo; ch++) {
      if (bits > extra_bits) add_bits[ch] = (extra_bits*add_bits[ch])/bits;
      targ_bits[ch] = targ_bits[ch] + add_bits[ch];
    }
    for (ch=0; ch<stereo; ch++) 
      extra_bits -= add_bits[ch];
  }

  if (reduce_sidechannel) {
    /*  ms_ener_ratio = 0:  allocate 66/33  mid/side  fac=.33  
     *  ms_ener_ratio =.5:  allocate 50/50 mid/side   fac= 0 */
    /* 75/25 split is fac=.5 */
    /* float fac = .50*(.5-ms_ener_ratio[gr])/.5;*/
    float fac = .33*(.5-ms_ener_ratio[gr])/.5;
    if (fac<0) fac=0;

    /* dont reduce side channel below 125 bits */
    if (targ_bits[1]-targ_bits[1]*fac > 125) {
      targ_bits[0] += targ_bits[1]*fac;
      targ_bits[1] -= targ_bits[1]*fac;
    }

  
    /* dont allow to many bits per channel */  
    for (ch=0; ch<stereo; ch++) {
      int max_bits = mean_bits/2 + 1200;
      if (targ_bits[ch] > max_bits) {
	extra_bits += (targ_bits[ch] - max_bits);
	targ_bits[ch] = max_bits;
      }
    }
  }  
  
  
  /* BEGIN MAIN LOOP */
  iteration = 0;
  compute_stepsize=1;

  while ( (notdone[0] || notdone[1])  ) {
    int pre_just_turned_on[2];
    iteration ++;

    if (compute_stepsize) {
      /* compute initial quantization step */
      compute_stepsize=0;
      for (ch=0 ; ch < stereo ; ch ++ )
	if (notdone[ch]) {
	  for(i=0;i<576;i++) 	    {
	    temp=fabs(xr[gr][ch][i]);
	    xrpow[gr][ch][i]=sqrt(sqrt(temp)*temp);
	  }
	  bits_found[ch]=bin_search_StepSize2(targ_bits[ch],-211.0,46,
	      l3_enc[gr][ch],xr[gr][ch],xrpow[gr][ch],cod_info[ch]); 
	}
    }



    /* inner_loop starts with the initial quantization step computed above
     * and slowly increases until the bits < huff_bits.
     * Thus is it important not to start with too large of an inital
     * quantization step.  Too small is ok, but inner_loop will take longer 
     */
    for (ch=0 ; ch < stereo ; ch ++ ) {
      if (notdone[ch]) {
	cod_info[ch]->part2_length = part2_length( scalefac, fr_ps, gr, ch, l3_side );
	huff_bits = targ_bits[ch] - cod_info[ch]->part2_length;
	if (huff_bits < 0) {
	  if (iteration==1) {
	    fprintf(stderr,"ERROR: outer_loop(): huff_bits < 0. \n");
	    exit(-5);
	  }else{
	    /* scale factors too large, not enough bits. use previous quantizaton */
	    notdone[ch]=0;
	    over[ch]=999;
	  }
	}else{
	  /* if this is the first iteration, see if we can reuse the quantization
	   * computed in bin_search_StepSize above */
	  if (iteration==1) {
	    if(bits_found[ch]>huff_bits) {
	      cod_info[ch]->quantizerStepSize+=1.0;
	      real_bits[ch] = inner_loop( xr, xrpow[gr][ch], l3_enc, huff_bits, cod_info[ch], gr, ch );
	    } else real_bits[ch]=bits_found[ch];
	  }
	  else 
	    real_bits[ch]=inner_loop( xr, xrpow[gr][ch], l3_enc, huff_bits, cod_info[ch], gr, ch );
	}
      }
    }



    /* compute the distortion in this quantization */
    if (fast_mode) {
      for (ch=0; ch<stereo; ch++)
	over[ch]=0;
    }else{
      if (convert_psy) {
	/* mid/side coefficiets, l/r thresholds */
	calc_noise2( xr[gr], l3_enc[gr], cod_info, xfsf,
          distort, l3_xmin,gr,stereo,over,over_noise,tot_noise,max_noise);
      }	else {
	  /* coefficients and thresholds both l/r (or both mid/side) */
	  for (ch=0; ch<stereo; ch++)
	    if (notdone[ch])
	      over[ch]=calc_noise1( xr[gr][ch], l3_enc[gr][ch], cod_info[ch], 
	    xfsf[ch],distort[ch], l3_xmin,gr,ch, &over_noise[ch], 
            &tot_noise[ch], &max_noise[ch]);
      }
    }

    /* check if this quantization is better the our saved quantization */
    if (iteration == 1) 
      for (ch=0; ch<stereo; ch++) better[ch]=1;
    else quant_compare_dual(better,notdone,stereo,ms_ener_ratio[gr],
        best_over,best_tot_noise,best_over_noise,best_max_noise,
	over,tot_noise,over_noise,max_noise);

 

    /* save data so we can restore this quantization later */    
    for (ch=0 ; ch < stereo ; ch ++ ) {
      if (better[ch]) {
	best_over[ch]=over[ch];
	best_over_noise[ch]=over_noise[ch];
	best_tot_noise[ch]=tot_noise[ch];
	best_max_noise[ch]=max_noise[ch];
	if (notdone[ch]) {
	  for ( sfb = 0; sfb < CBLIMIT; sfb++ ) /* save scaling factors */
	    scalesave_l[ch][sfb] = scalefac->l[gr][ch][sfb];
	  
	  for ( sfb = 0; sfb < SFB_SMAX; sfb++ )
	    for ( i = 0; i < 3; i++ )
	      scalesave_s[ch][sfb][i] = scalefac->s[gr][ch][sfb][i];
	  
	  save_preflag[ch]  = cod_info[ch]->preflag;
	  save_compress[ch] = cod_info[ch]->scalefac_compress;
	  
	  memcpy(save_l3_enc[ch],l3_enc[gr][ch],sizeof(l3_enc[gr][ch]));   
	  memcpy(&save_cod_info[ch],cod_info[ch],sizeof(save_cod_info[ch]));
	  save_real_bits[ch]=real_bits[ch];

#ifdef HAVEGTK
	  if (gtkflag) {
	    for ( i = 0; i < 3; i++ ) {
	      for ( sfb = cod_info[ch]->sfb_smax; sfb < 12; sfb++ )  {
		pinfo->xfsf_s[gr][ch][3*sfb+i] =  
		  pinfo->thr_s[gr][ch][3*sfb+i]*xfsf[ch][i+1][sfb]/
		  (1e-20+l3_xmin->s[gr][ch][sfb][i]);
	      }
	    }
	    for ( sfb = 0; sfb < cod_info[ch]->sfb_lmax; sfb++ )   {
	      pinfo->xfsf[gr][ch][sfb] =  
		pinfo->thr[gr][ch][sfb]*xfsf[ch][0][sfb]/
		(1e-20 + l3_xmin->l[gr][ch][sfb]);
	    }
	    pinfo->over[gr][ch]=over[ch];
	    pinfo->over_noise[gr][ch]=over_noise[ch];
	  }
#endif
	  

	}
      }
    }

    /* if no bands with distortion, we are done */
    for (ch=0 ; ch < stereo ; ch ++ ) 
      if (notdone[ch]) {
	if (convert_psy) 
	  notdone[ch] = (over[0] || over[1]);
	else
	  notdone[ch] = over[ch];
      }




    /* see if we should apply preemphasis */
    for (ch=0 ; ch < stereo ; ch ++ ) {
      pre_just_turned_on[ch]=0;
      if (notdone[ch] ) pre_just_turned_on[ch]=
	 preemphasis( xr[gr][ch], xrpow[gr][ch], l3_xmin, 
	       gr, ch, l3_side,	distort[ch]);
    }
    
    
    /* if we didn't just apply pre-emph, let us see if we should 
     * amplify some scale factor bands */
    for (ch=0 ; ch < stereo ; ch ++ ) 
      if (notdone[ch] && (!pre_just_turned_on[ch]) ) {
	  amp_scalefac_bands( xr[gr][ch], xrpow[gr][ch], l3_xmin,
		l3_side, scalefac, gr, ch, iteration,distort[ch]);
      }
    

    /* check to make sure we have not amplified too much */
    for (ch=0 ; ch < stereo ; ch ++ ) {
      if (notdone[ch]) {
	if ( (status[ch] = loop_break(scalefac, cod_info[ch], gr, ch)) == 0 ) {
	  if ( fr_ps->header->version == 1 ) {
	    status[ch] = scale_bitcount( scalefac, cod_info[ch], gr, ch );
	  }else{
	    status[ch] = scale_bitcount_lsf( scalefac, cod_info[ch], gr, ch );
	  }
        }
	notdone[ch] = !status[ch];
      }
    }
  }    /* done with main iteration */
  

  
  /* restore some data */
  for (ch=0 ; ch < stereo ; ch ++ ) {
    if (count[ch] ) {
      cod_info[ch]->preflag = save_preflag[ch];
      cod_info[ch]->scalefac_compress = save_compress[ch];
      
      for ( sfb = 0; sfb < CBLIMIT; sfb++ ) {
	scalefac->l[gr][ch][sfb] = scalesave_l[ch][sfb];    
      }
      
      for ( i = 0; i < 3; i++ )
	for ( sfb = 0; sfb < SFB_SMAX; sfb++ ) {
	  scalefac->s[gr][ch][sfb][i] = scalesave_s[ch][sfb][i];    
	}

      { 
	real_bits[ch]=save_real_bits[ch];  
	memcpy(l3_enc[gr][ch],save_l3_enc[ch],sizeof(l3_enc[gr][ch]));   
	memcpy(cod_info[ch],&save_cod_info[ch],sizeof(save_cod_info[ch]));
	
	if ( fr_ps->header->version == 1 )
	  status[ch] = scale_bitcount( scalefac, cod_info[ch], gr, ch );
	else
	  status[ch] = scale_bitcount_lsf( scalefac, cod_info[ch], gr, ch );
	if (status[ch]) {
	  fprintf(stderr,"Error recomputing scalefac_compress...this should not happen");
	  exit(-10);
	}
      }
      cod_info[ch]->part2_length   = part2_length( scalefac, fr_ps, gr, ch, l3_side );
      cod_info[ch]->part2_3_length = cod_info[ch]->part2_length + real_bits[ch];

#ifdef HAVEGTK
      if (gtkflag)
	pinfo->LAMEmainbits[gr][ch]=cod_info[ch]->part2_3_length;
#endif
    }      
  }
  
  /* finish up */
  for (ch=0 ; ch < stereo ; ch ++ ) {
    if (!VBR) ResvAdjust( fr_ps, cod_info[ch], l3_side, mean_bits );
    cod_info[ch]->global_gain = nint( cod_info[ch]->quantizerStepSize + 210.0 );
    assert( cod_info[ch]->global_gain < 256 );
  }
}



