/**********************************************************************
 * ISO MPEG Audio Subgroup Software Simulation Group (1996)
 * ISO 13818-3 MPEG-2 Audio Encoder - Lower Sampling Frequency Extension
 *
 * $Id: musicin.c,v 1.2 1997/01/19 22:28:29 rowlands Exp $
 *
 * $Log: musicin.c,v $
 * Revision 1.2  1997/01/19 22:28:29  rowlands
 * Layer 3 bug fixes from Seymour Shlien
 *
 * Revision 1.1  1996/02/14 04:04:23  rowlands
 * Initial revision
 *
 * Received from Mike Coleman
 **********************************************************************/
#include <sys/types.h>
#include <sys/stat.h>
#ifdef HAVETIMES
#include <sys/times.h>
#endif
#include <fcntl.h>
#include <stdlib.h>
#include <time.h>
#include <assert.h>
#ifdef _WIN32
#include <io.h> 
#endif

#include "musicin.h"
#include "l3psy.h"
#include "mdct.h"
#include "loop.h"
#include "l3bitstream.h"
#ifdef HAVEGTK
#include "gtkanal.h"
#include <gtk/gtk.h>
#endif
#include "version.h"
#include "VbrTag.h"
#include "get_audio.h"



/* Global flags.  set defaults here, defined extern in common.h */
int autoconvert=FALSE;
int force_ms=FALSE;
int gpsycho=1;
int sfb21=1;
int experimentalX = 0;
int experimentalY = 0;
int experimentalZ = 0;
int VBR=0;
int VBR_q=4;
int VBR_min_bitrate=1;   /* 32kbs */
int VBR_max_bitrate=13;  /* 256kbs */
int silent=0;
int highq=0;
int fast_mode=0;
int allow_diff_short=0;
#ifdef DISTRIB
int gtkflag=1;
#else
int gtkflag=0;
#endif
int g_bWriteVbrTag=1;
int iswav=0;
int ispcm=0;
long totalframes;        /* frames: 0..totalframes-1 */
int original_file_type;  /* 0=WAV, 1=MP3 */
int frameNum=0;


/* Global variable definitions for "musicin.c" */
struct stat sb;
Bit_stream_struc   bs;
III_side_info_t l3_side;
frame_params fr_ps;
char    inPath[MAX_NAME_SIZE];
char    outPath[MAX_NAME_SIZE];
int target_bitrate;
int swapbytes = FALSE; /* force byte swapping */
char               *programName;
unsigned long num_samples;
#ifdef HAVEGTK
static int MPGLAG=1;
#endif
int original_file_fd;    /* file descriptor */
static layer info;


#ifdef BRHIST
#include <string.h>
#include <termcap.h>
void brhist_init(int br_min, int br_max);
void brhist_add_count(void);
void brhist_disp(void);
void brhist_disp_total(void);
extern long brhist_temp[15];

int disp_brhist = 1;
#endif



/*********************************************************/
/* timestatus: display encoding process time information */
/*********************************************************/
void timestatus()
{
  static double time_so_far=0.;
  static clock_t clock_value=0;
  clock_t last_clock_value;
  double estimated, speedup;
  int samp_rate=s_freq[info.version][info.sampling_frequency];
  static double process_time_so_far=0.;
  static clock_t process_clock_value=0;
  clock_t process_last_clock_value;
  double process_estimated, process_speedup;
  int percent;

#ifdef HAVETIMES
  struct tms tms_buf;
#endif
  int eta;
  char str2[31];

#ifdef CLK_TCK
  static int divider=CLK_TCK;
#else
  static int divider=HZ;         /* for old BSD systems */
#endif

  last_clock_value=clock_value;
  process_last_clock_value=process_clock_value;

#ifdef HAVETIMES
  clock_value=times(&tms_buf);   /* wall clock */
#else
  clock_value=clock();           /* is there a Mac/Windows wall clock function?? */
#endif
  process_clock_value=clock();   /* cpu time of process */

  if (frameNum==0) {
    last_clock_value=clock_value;
    process_last_clock_value=process_clock_value;
    fprintf(stderr,"    Frame          |  CPU/estimated  |  time/estimated | play/CPU |   ETA\n");
  }

  /* calculate both elapsed time and time used by process in seconds */
  time_so_far+=(double)(clock_value-last_clock_value)/(double)divider;
  process_time_so_far+=(double)(process_clock_value-process_last_clock_value)/(double)CLOCKS_PER_SEC;

  /* if we haven't encoded any frames we won't show any encoding status yet */
  if ((frameNum==0) || process_time_so_far==0.) return;

  /* calc estimated total CPU time and estimated time */
  estimated = (double)totalframes/(double)frameNum*time_so_far;
  process_estimated = (double)totalframes/(double)frameNum*process_time_so_far;

  /* playing time / estimated encoding CPU time. */
  if (samp_rate) {
    /*speedup = (double)totalframes*1.152/((double)samp_rate*estimated);*/
    process_speedup = (double)totalframes*1.152/((double)samp_rate*process_estimated);
  }
  else {
    process_speedup = speedup = 0.;
  }

  sprintf(str2,"%.4f",process_speedup);
  if (strlen(str2)<9) memset(str2+strlen(str2), ' ', 13-strlen(str2));
  str2[9]=0;
  
#ifdef HAVETIMES
  eta = estimated-time_so_far;
#else
  eta = process_estimated-process_time_so_far;
#endif

  percent = (int)(100.0*( (float)frameNum/(float)(totalframes-1) ));

  fprintf(
    stderr,
    /*"Frame: %d/%d CPU time: %d:%02d:%02d/%d:%02d:%02d (%.4f) Time: %d:%02d:%02d/%d:%02d:%02d  \r",*/
    "\r%6d/%6d(%3d%%)|%2d:%02d:%02d/%2d:%02d:%02d|%2d:%02d:%02d/%2d:%02d:%02d| %s|%2d:%02d:%02d ",
    frameNum, 
    (int)totalframes-1,
    percent,
    ((int)process_time_so_far/3600), 
    ((int)(process_time_so_far/60)%60), 
    (nint(process_time_so_far)%60),
    ((int)process_estimated/3600),
    (((int)process_estimated/60)%60), 
    (nint(process_estimated)%60),
    (((int)time_so_far/3600)), 
    (((int)time_so_far/60)%60), 
    (nint(time_so_far)%60),
    (((int)estimated/3600)),(((int)estimated/60)%60), 
    (nint(estimated)%60),
    str2,
    (((int)eta/3600)%24),(((int)eta/60)%60), ((int)eta%60));

  fflush(stderr);
  
}



/************************************************************************
*
* usage
*
* PURPOSE:  Writes command line syntax to the file specified by #stderr#
*
************************************************************************/

void lame_usage(char *name)  /* print syntax & exit */
{
  fprintf(stderr,"LAME Ain't an Mp3 Encoder (Reconstruction-iso)\n");
  fprintf(stderr,"LAME version %s (www.sulaco.org/mp3) \n",get_lame_version());
  fprintf(stderr,"GPSYCHO: GPL psycho-acoustic model version %s. \n\n",get_psy_version());
  fprintf(stderr,"USAGE   :  %s [options] <infile> <outfile>\n",name);
  fprintf(stderr,"OPTIONS :\n");
  fprintf(stderr,"          -m mode         (s)tereo, (j)oint, (f)orce or (m)ono  (default %c)\n",DFLT_MOD);
  fprintf(stderr,"                          force = force ms_stereo on all frames. Faster and\n");
  fprintf(stderr,"                          uses special Mid & Side masking thresholds\n");
  fprintf(stderr,"          -s sample freq  frequency of input file (kHz) - default %4.1f\n",DFLT_SFQ);
  fprintf(stderr,"          -b <bitrate>    set the bitrate, default 128kbps\n");
  fprintf(stderr,"                          (for VBR, this sets the allowed minimum bitrate)\n");
  fprintf(stderr,"\n");
  fprintf(stderr,"          -v              use variable bitrate (VBR)\n");
  fprintf(stderr,"          -V n            quality setting for VBR.  default n=%i\n",VBR_q);
  fprintf(stderr,"                          0=high quality,bigger files. 9=smaller files\n");
  fprintf(stderr,"          -t              disable Xing VBR informational tag\n");
  fprintf(stderr,"          --nohist        disable VBR histogram display\n");
  fprintf(stderr,"\n");
  fprintf(stderr,"          -h              use (maybe) quality improvements\n");
  fprintf(stderr,"          -f              fast mode (low quality)\n");
  fprintf(stderr,"          -k              disable sfb=21 cutoff\n");
  fprintf(stderr,"          -d              allow channels to have different blocktypes\n");
  fprintf(stderr,"\n");
#ifndef LIBSNDFILE
  fprintf(stderr,"          -r              input is raw pcm\n");
#endif
  fprintf(stderr,"          -x              force byte-swapping of input\n");
  fprintf(stderr,"          -a              downmix from stereo to mono file for mono encoding\n");
  fprintf(stderr,"          -e emp          de-emphasis n/5/c        (default %4c)\n",DFLT_EMP);
  fprintf(stderr,"          -p              add CRC error protection (now working!)\n");
  fprintf(stderr,"          -c              mark as copyright\n");
  fprintf(stderr,"          -o              mark as non-original\n");
  fprintf(stderr,"          -S              don't print progress report, VBR histograms\n");
  fprintf(stderr,"\n");
#ifdef HAVEGTK
  fprintf(stderr,"          -g              run graphical analysis on <infile>\n");
#endif
  fprintf(stderr,"\n<infile> and/or <outfile> can be \"-\", which means stdin/stdout.\n");

  exit(1);
}






/************************************************************************
*
* parse_args
*
* PURPOSE:  Sets encoding parameters to the specifications of the
* command line.  Default settings are used for parameters
* not specified in the command line.
*
* If the input file is in WAVE or AIFF format, the sampling frequency is read
* from the AIFF header.
*
* The input and output filenames are read into #inpath# and #outpath#.
*
************************************************************************/
void
lame_parse_args(argc, argv)
int     argc;
char    **argv;
{
  FLOAT srate;
  int   brate;
  layer *info = fr_ps.header;
  int   err = 0, i = 0;
  int mode_given=0;
  int num_channels=2;   /* number of channels of INPUT file */
  long samplerate;

  /* preset defaults */
  programName = argv[0]; 
  inPath[0] = '\0';   outPath[0] = '\0';
  info->lay = DFLT_LAY;
  switch(DFLT_MOD) {
  case 's': info->mode = MPG_MD_STEREO; info->mode_ext = 0; break;
  case 'd': info->mode = MPG_MD_DUAL_CHANNEL; info->mode_ext=0; break;
  case 'j': info->mode = MPG_MD_JOINT_STEREO; info->mode_ext=0; break;
  case 'm': info->mode = MPG_MD_MONO; info->mode_ext = 0; break;
  default:
    fprintf(stderr, "%s: Bad mode dflt %c\n", programName, DFLT_MOD);
    abort();
  }
  if((info->sampling_frequency = SmpFrqIndex((long)(1000*DFLT_SFQ), &info->version)) < 0) {
    fprintf(stderr, "%s: bad sfrq default %.2f\n", programName, DFLT_SFQ);
    abort();
  }
  info->bitrate_index = 9;
  brate = 0;
  switch(DFLT_EMP) {
  case 'n': info->emphasis = 0; break;
  case '5': info->emphasis = 1; break;
  case 'c': info->emphasis = 3; break;
  default: 
    fprintf(stderr, "%s: Bad emph dflt %c\n", programName, DFLT_EMP);
    abort();
  }
  info->copyright = 0;
  info->original = 1;
  info->error_protection = FALSE;
  
  /* process args */
  while(++i<argc && err == 0) {
    char c, *token, *arg, *nextArg;
    int  argUsed;
    
    token = argv[i];
    if(*token++ == '-') {
      if(i+1 < argc) nextArg = argv[i+1];
      else           nextArg = "";
      argUsed = 0;
      if (! *token) {
	/* The user wants to use stdin and/or stdout. */
	if(inPath[0] == '\0')       strncpy(inPath, argv[i],MAX_NAME_SIZE);
	else if(outPath[0] == '\0') strncpy(outPath, argv[i],MAX_NAME_SIZE);
      } 
      if (*token == '-') {
	/* GNU style */
	token++;
#ifdef BRHIST
	if (strcmp(token, "nohist")==0)
	  disp_brhist = 0;
	else
#endif
	  {
	    fprintf(stderr,"%s: unrec option --%s\n",
		    programName, token);
	  }
	continue;
      }
      while( (c = *token++) ) {
	if(*token ) arg = token;
	else                             arg = nextArg;
	switch(c) {
	case 'm':        argUsed = 1;
	  mode_given=1;
	  if (*arg == 's')
	    { info->mode = MPG_MD_STEREO; info->mode_ext = 0; }
	  else if (*arg == 'd')
	    { info->mode = MPG_MD_DUAL_CHANNEL; info->mode_ext=0; }
	  else if (*arg == 'j')
	    { info->mode = MPG_MD_JOINT_STEREO; info->mode_ext=0; }
	  else if (*arg == 'f')
	    { info->mode = MPG_MD_JOINT_STEREO; force_ms=1; info->mode_ext=0; }
	  else if (*arg == 'm')
	    { info->mode = MPG_MD_MONO; info->mode_ext = 0; }
	  else {
	    fprintf(stderr,"%s: -m mode must be s/d/j/f/m not %s\n",
		    programName, arg);
	    err = 1;
	  }
	  break;
	case 'V':        argUsed = 1;   VBR = 1;
	  if (*arg == '0')
	    { VBR_q=0; }
	  else if (*arg == '1')
	    { VBR_q=1; }
	  else if (*arg == '2')
	    { VBR_q=2; }
	  else if (*arg == '3')
	    { VBR_q=3; }
	  else if (*arg == '4')
	    { VBR_q=4; }
	  else if (*arg == '5')
	    { VBR_q=5; }
	  else if (*arg == '6')
	    { VBR_q=6; }
	  else if (*arg == '7')
	    { VBR_q=7; }
	  else if (*arg == '8')
	    { VBR_q=8; }
	  else if (*arg == '9')
	    { VBR_q=9; }
	  else {
	    fprintf(stderr,"%s: -V n must be 0-9 not %s\n",
		    programName, arg);
	    err = 1;
	  }
	  break;
	  
	case 's':
	  argUsed = 1;
	  srate = atof( arg );
	  /* samplerate = rint( 1000.0 * srate ); $A  */
	  samplerate = (long) (( 1000.0 * srate ) + 0.5);
	  if( (info->sampling_frequency =
	       SmpFrqIndex((long) samplerate, &info->version)) < 0 )
	    err = 1;
	  break;
	  
	case 'b':        
	  argUsed = 1;
	  brate = atoi(arg); 
	  break;
	case 't':  /* dont write VBR tag */
		g_bWriteVbrTag=0;
	  break;
	case 'r':  /* force raw pcm input file */
#ifdef LIBSNDFILE
	  fprintf(stderr,
	  "%s compiled with LIBSNDFILE.  Compile without LIBSNDFILE to use -r\n",
	  programName);
	  exit(1);
#else
	  ispcm=TRUE;
#endif	
	  break;
	case 'x':  /* force byte swapping */
	  swapbytes=TRUE;
	  break;
	case 'p': /* (jo) error_protection: add crc16 information to stream */
	  info->error_protection = 1; 
	  break;
	case 'a': /* autoconvert input file from stereo to mono - for mono mp3 encoding */
	  autoconvert = TRUE;
	  info->mode = MPG_MD_MONO; 
	  info->mode_ext = 0;
	  break;
	case 'h': 
	  highq = TRUE;
	  break;
	case 'k': 
	  sfb21 = FALSE;
	  break;
	case 'd': 
	  allow_diff_short = 1;
	  break;
	case 'v': 
	  VBR = 1;
	  break;
	case 'S': 
	  silent = TRUE;
	  break;
	case 'X':        argUsed = 1;   experimentalX = 0;
	  if (*arg == '0')
	    { experimentalX=0; }
	  else if (*arg == '1')
	    { experimentalX=1; }
	  else if (*arg == '2')
	    { experimentalX=2; }
	  else if (*arg == '3')
	    { experimentalX=3; }
	  else if (*arg == '4')
	    { experimentalX=4; }
	  else if (*arg == '5')
	    { experimentalX=5; }
	  else {
	    fprintf(stderr,"%s: -X n must be 0-5 not %s\n",
		    programName, arg);
	    err = 1;
	  }
	  break;


	case 'Y': 
	  experimentalY = TRUE;
	  break;
	case 'Z': 
	  experimentalZ = TRUE;
	  break;
	case 'f': 
	  fast_mode = TRUE;
	  break;
#ifdef HAVEGTK
	case 'g': /* turn on gtk analysis */
	  gtkflag = TRUE;
	  break;
#endif
	case 'e':        argUsed = 1;
	  if (*arg == 'n')                    info->emphasis = 0;
	  else if (*arg == '5')               info->emphasis = 1;
	  else if (*arg == 'c')               info->emphasis = 3;
	  else {
	    fprintf(stderr,"%s: -e emp must be n/5/c not %s\n",
		    programName, arg);
	    err = 1;
	  }
	  break;
	case 'c':       info->copyright = 1; break;
	case 'o':       info->original  = 0; break;
	default:        fprintf(stderr,"%s: unrec option %c\n",
				programName, c);
	err = 1; break;
	}
	if(argUsed) {
	  if(arg == token)    token = "";   /* no more from token */
	  else                ++i;          /* skip arg we used */
	  arg = ""; argUsed = 0;
	}
      }
    }
    else {
      if(inPath[0] == '\0')       strcpy(inPath, argv[i]);
      else if(outPath[0] == '\0') strcpy(outPath, argv[i]);
      else {
	fprintf(stderr,"%s: excess arg %s\n", programName, argv[i]);
	err = 1;
      }
    }
  }

  // Do not write VBR tag if VBR flag is not specified
  if (g_bWriteVbrTag==1 && VBR==0) g_bWriteVbrTag=0;

  if(err || inPath[0] == '\0') lame_usage(programName);  /* never returns */
  if (inPath[0]=='-') silent=1;  /* turn off status - it's broken for stdin */
  if(outPath[0] == '\0') {
    if (inPath[0]=='-') {
      /* if input is stdin, default output is stdout */
      strcpy(outPath,"-");
    }else {
      strcpy(outPath, inPath);
      strcat(outPath, DFLT_EXT);
    }
  }
  if (outPath[0]=='-') g_bWriteVbrTag=0; /* turn off VBR tag */
  
  // SPECIAL FOR BLADE ENCODER
#ifdef _BLADEDLL
  open_bit_stream_w(&bs, outPath, BUFFER_SIZE);
#else
  // are we encoding to .mp3, or analyzing a .mp3 file?
  original_file_type=!(strcmp((char *) &inPath[strlen(inPath)-4],".mp3"));
  
  if (original_file_type)    {
    /* MP3 file.  */
    totalframes = 0;
    original_file_fd = open(inPath,O_RDONLY);
    if (original_file_fd == -1) {
      fprintf(stderr,"Could not find \"%s\".\n", inPath);
      exit(1);
    }
    if (!gtkflag) {
      fprintf(stderr,"Use option -g to analyze mp3 files.\n");
      exit(1);
    }
     
  }else{
#ifdef DISTRIB
    outPath="/dev/null";
#endif
    open_bit_stream_w(&bs, outPath, BUFFER_SIZE);  /* open the output file */
    OpenSndFile(inPath,info);                      /* open input file */

    // Set sample frequency
    info->sampling_frequency = SmpFrqIndex(GetSndSampleRate(), &info->version);
    num_channels=GetSndChannels();
    if (num_channels==1)
      {
	autoconvert = FALSE;  /* avoid 78rpm emulation mode from downmixing mono file! */
	info->mode = MPG_MD_MONO;
	info->mode_ext = 0;
      }
    num_samples = GetSndSamples();
    if ( (num_samples!=MAX_U_32_NUM)) num_samples *= num_channels;

  } // MP3 or PCM file?
#endif   /* BLADEDLL*/



  if (g_bWriteVbrTag)
    {
      // Write initial VBR Header to bitstream
      InitVbrTag(&bs,info->version-1,info->mode,info->sampling_frequency);
    }


  /* Timing, autoconvert and resampling checks and setup */
  /* dont do this if analyzing a mp3 file */
  if (!original_file_type)    {
    int framesize = (info->version==0) ? 576 : 1152;

    if (num_samples == MAX_U_32_NUM) { 
      stat(inPath,&sb);  /* try file size, assume 2 bytes per sample */
      totalframes = 2+(sb.st_size/(2*framesize*num_channels));
    }else{
      totalframes = 2+(num_samples/(framesize*num_channels));
    }
    if (autoconvert==TRUE) {
      fprintf(stderr, "Autoconverting from stereo to mono. Setting encoding to mono mode.\n");
    }
  }
  
  
  if ( brate == 0 ) {
    brate = bitrate[info->version][info->lay-1][9];
    info->bitrate_index=9;
  } else {
    if( (info->bitrate_index = BitrateIndex(info->lay, brate, info->version)) < 0)
      err=1;
    /* a bit rate was specified.  for VBR, take this to be the minimum */
    VBR_min_bitrate=info->bitrate_index;
  }
  if ((!mode_given) && (brate >= 192)) {
    /* default is STEREO for higher bitrates, not JSTEREO */
    if (info->mode != MPG_MD_MONO){
      info->mode = MPG_MD_STEREO; info->mode_ext = 0;
    }
  }

  if(err || inPath[0] == '\0') lame_usage(programName);  /* never returns */
  if (info->bitrate_index==14) VBR=0;  /* dont bother with VBR at 320kbs */
  /* we do not normally allow 320bps frams with VBR, unless: */
  if (info->bitrate_index==13) VBR_max_bitrate=14;  
  if (VBR_q == 0) VBR_max_bitrate=14;  
  if (VBR) {
    fast_mode = FALSE; 
    highq = TRUE;
  }

  /* dont allow forced mid/side stereo for mono signals */
  if (info->mode == MPG_MD_MONO) {
    force_ms=0;  
  }

  /* MPEG2 specific requirements */
  if (info->version==0) {
    g_bWriteVbrTag=0;      /* no MPEG2 Xing VBR tags yet */
  }

  /* copy some header information */
  hdr_to_frps(&fr_ps);

#ifdef BRHIST
  if (VBR) {
    if (disp_brhist)
      brhist_init(VBR_min_bitrate, VBR_max_bitrate);
  } else
    disp_brhist = 0;
#endif
}




/************************************************************************
 *
 * print_config
 *
 * PURPOSE:  Prints the encoding parameters used
 *
 ************************************************************************/
void lame_print_config()
{
  layer *info = fr_ps.header;
  if (original_file_type) {
    fprintf(stderr, "Analyzing %s \n",inPath);
  }
  else {
    fprintf(stderr, "Encoding %s to %s\n",
	    (strcmp(inPath, "-")? inPath : "stdin"),
	    (strcmp(outPath, "-")? outPath : "stdout"));
    if (VBR)
      fprintf(stderr, "Encoding as %.1f kHz VBR(q=%i) %s MPEG%i LayerIII file\n",
	      s_freq[info->version][info->sampling_frequency],
	      VBR_q,mode_names[info->mode],2-info->version);
    else
      fprintf(stderr, "Encoding as %.1f kHz %d kbps %s MPEG%i LayerIII file\n",
	      s_freq[info->version][info->sampling_frequency],
	      bitrate[info->version][info->lay-1][info->bitrate_index],
	      mode_names[info->mode],2-info->version);
  }
  fflush(stderr);
}

#ifdef HAVEGTK
int lame_decode_fromfile()
{
  int j,ch,stereo=2,out=0;
  static int adj=0;
  short mpg123pcm[2][1152];

  /* create a 1 frame lag so MP3 analysis syncs with WAV analysis */
  //  if (frameNum == 0 ) out=1152;
  //  else 
  out=decode1file(original_file_fd,mpg123pcm,&stereo);

  /* are the frames 1152 or 576 samples? */
  if (out) {
    adj=0;
    if (out==576) adj=576;
    fr_ps.stereo=stereo;

    /* Compensate for the decoder_delay.  add DECDELAY, where
     * decoder_delay + DECDELAY = 1152 (or 576), so data is delayed
     * by exactly one frame */
    /* if out=0, last frame of data is in mpg123pcm */
    for ( ch = 0; ch < stereo; ch++ ) {
      for ( j = 0; j < DECDELAY-adj; j++ )
	pinfo->pcmdata2[ch][j] = pinfo->pcmdata2[ch][j+1152-adj];
      for ( j = 0; j < 1152-adj; j++ ) 
	pinfo->pcmdata2[ch][j+DECDELAY-adj] = mpg123pcm[ch][j];
    }
  }
  
  MPGLAG = 1;
  pinfo->mpglag = MPGLAG;
  pinfo->frameNum = frameNum;
  frameNum++;
  return out;
}
#endif



/************************************************************************
* 
* encodeframe()
*
* SEMANTICS:  One overlapping frame of audio of up to 2 channels are
* processed at a time in the following order:
* (associated routines are in parentheses)
*
* 1.  Filter sliding window of data to get 32 subband
* samples per channel.
* (window_subband,filter_subband)
*
* 2.  Calculate scalefactors for the frame, and if layer 2,
* also calculate scalefactor select information.
* (*_scale_factor_calc)
*
* 3.  Calculate psychoacoustic masking levels using selected
* psychoacoustic model.
* (*_Psycho_One, psycho_anal)
*
* 4.  Perform iterative bit allocation for subbands with low
* mask_to_noise ratios using masking levels from step 4.
* (*_main_bit_allocation)
*
* 5.  Pack bit allocation and scalefactors onto bitstream.
* (*_encode_bit_alloc,*_encode_scale)
*
* 6.  Quantize subbands and pack them into bitstream
* (*_subband_quantization, *_sample_encoding)
*
************************************************************************/
int lame_encode_tofile(short int Buffer[2][1152])
{
  static unsigned long frameBits;
  static unsigned long bitsPerSlot;
  static double frac_SpF;
  static double slot_lag;
  static int mode_gr;
  static int mpg123count=0;
  static int old_bitrate;
  static unsigned long sentBits = 0;
  static L3SBS     l3_sb_sample;
  /* set to 56 to sync with FhG */
#define EXTRADELAY 56
  static int samplesPerFrame;
  static short int mfbuf[2][1152+576+EXTRADELAY];   /* 3 granule internal buffer */
  double xr[2][2][576];
  double win_que[64];
  int l3_enc[2][2][576];

  III_psy_ratio ratio;
  III_scalefac_t scalefac;

  int j,ch,gr,mean_bits;
  int bitsPerFrame;
  double pe[2][4];
  double pe_use[2][2];
  double pe_sum_old=0;
  static double pe_sum=0;
  short *win_buf[2];
  int stereo;
  layer *info;
  int i;


#ifdef HAVEGTK
  char mpg123bs[BUFFER_SIZE];
  short mpg123pcm[2][1152];
#endif
  int check_ms_stereo;
  int bit_rate;
  double samp;
  double ms_ener_ratio[2];

  stereo = fr_ps.stereo;
  info = fr_ps.header;

#ifdef HAVEGTK
  if (gtkflag) {
    pinfo->frameNum = frameNum;
    pinfo->sampfreq=s_freq[info->version][info->sampling_frequency]*1000.0;
  }
#endif

  info->mode_ext = 0; 

  /* use m/s stereo? */
  check_ms_stereo =   ((info->mode == MPG_MD_JOINT_STEREO) && 
		       (!force_ms) && (info->version == 1) &&
		       (stereo==2));


  if (frameNum==0) old_bitrate=info->bitrate_index;
  info->bitrate_index = old_bitrate;


  bit_rate = bitrate[info->version][info->lay-1][info->bitrate_index];
  samp =      s_freq[info->version][info->sampling_frequency];
  
  if (frameNum==0)  {    
    double avg_slots_per_frame;
    int whole_SpF;
    mpg123count=0;
    sentBits = 0;
    memset((char *) mfbuf, 0, sizeof(mfbuf));

    /* Figure average number of 'slots' per frame. */
    /* Bitrate means TOTAL for both channels, not per side. */
    bitsPerSlot = 8;
    samplesPerFrame = info->version == 1 ? 1152 : 576;
    avg_slots_per_frame = ((double)samplesPerFrame /samp) *
      ((double)bit_rate /  (double)bitsPerSlot);
    whole_SpF = (int) avg_slots_per_frame;
    frac_SpF  = avg_slots_per_frame - (double)whole_SpF;
    slot_lag  = -frac_SpF;
    info->padding = 1;
    if (fabs(frac_SpF) < 1e-9) info->padding = 0;
    mode_gr = (info->version == 1) ? 2 : 1;  /* mode_gr = 2 */
  }


  /* shift in new samples, delayed by 576+EXTRADELAY */
   for (ch=0; ch<stereo; ch++)
     for (i=0; i<576+EXTRADELAY; i++)
       mfbuf[ch][i]=mfbuf[ch][i+samplesPerFrame];
   for (ch=0; ch<stereo; ch++)
     for (i=0; i<samplesPerFrame; i++)
       mfbuf[ch][i+576+EXTRADELAY]=Buffer[ch][i];




  /********************** padding *****************************/
  if (VBR) {
  /* leave info_padding as it was set above */
  } else {
    if (frac_SpF != 0) {
      if (slot_lag > (frac_SpF-1.0) ) {
	slot_lag -= frac_SpF;
	info->padding = 0;
      }
      else {
	info->padding = 1;
	slot_lag += (1-frac_SpF);
      }
    }
  }


  /********************** status display  *****************************/
  if (!gtkflag && !silent) {
    if (frameNum%20==0) {
      timestatus();
#ifdef BRHIST
      if (disp_brhist)
	{
	  brhist_add_count();
	  brhist_disp();
	}
#endif
    }
    if (frameNum==totalframes-1) {
      timestatus();
#ifdef BRHIST
      if (disp_brhist)
	{
	  brhist_add_count();
	  brhist_disp();
	}
#endif
      fprintf(stderr,"\n");
#ifdef BRHIST
      if (disp_brhist)
	brhist_disp_total();
#endif
      fflush(stderr);
    }
  }


  /********************** process InputBuffer *****************************/


  if (force_ms)
    for (i=0; i< samplesPerFrame; i++) {
      /* dont overflow the short int */
      double mid = ((int)mfbuf[0][576+i]+(int)mfbuf[1][576+i])/(2.0);
      double side= ((int)mfbuf[0][576+i]-(int)mfbuf[1][576+i])/(2.0);
      mfbuf[0][576+i]=mid;
      mfbuf[1][576+i]=side;
    }


  
  /***************************** Layer 3 **********************************
   * mfbuf contains 3 granules:  [0 1 2 ] 
   * encoder will encode granules 0 1 
   * psy-model will be fed granules 1 2, and because of its 1 granule delay 
   * it will return thresholds for granules 0 1 */
  
  win_buf[0] = &mfbuf[0][0];
  win_buf[1] = &mfbuf[1][0];

  
  if (!fast_mode) {  
    /* psychoacoustic model 
     * psy model adds a 544 delay to sync with the filterbanks
     * in addition to this delay, it also adds a 1 granule (576) delay
     * that we must compensate for (mt 6/99). 
     */
    short int *bufp[2];  /* address of beginning of left & right granule */
    int blocktype[2];
    
    for (gr=0; gr < mode_gr ; gr++) {
      for ( ch = 0; ch < stereo; ch++ ) bufp[ch]=&mfbuf[ch][576 + gr*576];
      L3psycho_anal( bufp, stereo, gr, info,
         s_freq[info->version][info->sampling_frequency] * 1000.0,
	 check_ms_stereo,&ms_ener_ratio[gr],
         ratio.l[gr], ratio.s[gr], pe[gr], blocktype);
      for ( ch = 0; ch < stereo; ch++ ) 
	l3_side.gr[gr].ch[ch].tt.block_type=blocktype[ch];
    }
  }else{
    for (gr=0; gr < mode_gr ; gr++) 
      for ( ch = 0; ch < stereo; ch++ ) {
	l3_side.gr[gr].ch[ch].tt.block_type=NORM_TYPE;
	pe[gr][ch]=700;
      }
  }

  pe_sum_old=pe_sum;
  pe_sum=0;
  for( gr = 0; gr < mode_gr; gr++ ) 
    for ( ch = 0; ch < stereo; ch++ ) 
      pe_sum+=pe[gr][ch];
  

  /* for VBR, what bitrate should we try first? */
  if (VBR){
    info->bitrate_index=old_bitrate-1;
    if (abs(pe_sum_old - pe_sum)>200){
      if (pe_sum_old>pe_sum)
        info->bitrate_index--;
    }

    if (info->bitrate_index < VBR_min_bitrate)
      info->bitrate_index=VBR_min_bitrate;
    if (info->bitrate_index > VBR_max_bitrate)
      info->bitrate_index=VBR_max_bitrate;
  }




  /* block type flags */
  for( gr = 0; gr < mode_gr; gr++ ) {
    for ( ch = 0; ch < stereo; ch++ ) {
      gr_info *cod_info = &l3_side.gr[gr].ch[ch].tt;
      cod_info->mixed_block_flag = 0;     /* never used by this model */
      if (cod_info->block_type == NORM_TYPE )
	cod_info->window_switching_flag = 0;
      else
	cod_info->window_switching_flag = 1;
    }
  }

  
  /* polyphase filtering  */
  for( gr = 0; gr < mode_gr; gr++ )
    for ( ch = 0; ch < stereo; ch++ )
      for ( j = 0; j < 18; j++ )   {
	window_subband( &win_buf[ch], win_que, ch );
	filter_subband( win_que,  &(l3_sb_sample)[ch][gr+1][j][0] ); }
  
  /* apply mdct to the polyphase outputs */
  mdct_sub( &l3_sb_sample, xr, stereo, &l3_side, mode_gr );


  /* data was scaled by 1/2.  fix so effectively it was scaled by 1/sqrt(2) */
  if (force_ms) {
    for ( gr = 0; gr < mode_gr; gr++ )
      for ( ch = 0; ch < stereo; ch++ ) 
	for (i =0 ; i< 576; i++) 
	  xr[gr][ch][i] *= sqrt(2.0);
  }


  if (check_ms_stereo) {
    /* make sure block type is the same in each channel */
    check_ms_stereo = 
      (l3_side.gr[0].ch[0].tt.block_type==l3_side.gr[0].ch[1].tt.block_type) &&
      (l3_side.gr[1].ch[0].tt.block_type==l3_side.gr[1].ch[1].tt.block_type);
  }
  if (check_ms_stereo) {
    if ( (ms_ener_ratio[0] + ms_ener_ratio[1])<.70) info->mode_ext = 2;
  }

#ifdef HAVEGTK
  if (gtkflag) { 
    for ( gr = 0; gr < mode_gr; gr++ )
      for ( ch = 0; ch < stereo; ch++ ) {
	pinfo->ms_ratio[gr]=ms_ener_ratio[gr];
	pinfo->blocktype[gr][ch]=
	  l3_side.gr[gr].ch[ch].tt.block_type;
	for ( j = 0; j < 576; j++ ) pinfo->xr[gr][ch][j]=xr[gr][ch][j];
      }
  }
#endif

  /* if MS stereo, switch to MS psy data */
  for ( gr = 0; gr < mode_gr; gr++ ) {
    for ( ch = 0; ch < stereo; ch++ ) {
      if (highq && (info->mode_ext==2)) {
	pe_use[gr][ch]=pe[gr][ch+2];
	memcpy(ratio.l[gr][ch],ratio.l[gr][ch+2],sizeof(ratio.l[gr][ch]));
	memcpy(ratio.s[gr][ch],ratio.s[gr][ch+2],sizeof(ratio.s[gr][ch]));
#ifdef HAVEGTK
       if (gtkflag) {
	 pinfo->ers[gr][ch]=pinfo->ers[gr][ch+2];
	 pinfo->pe[gr][ch]=pinfo->pe[gr][ch+2];
         memcpy(pinfo->energy[gr][ch],pinfo->energy[gr][ch+2],
		sizeof(pinfo->energy[gr][ch]));
	 memcpy(pinfo->thr[gr][ch],pinfo->thr[gr][ch+2],
		sizeof(pinfo->thr[gr][ch]));
	 memcpy(pinfo->en[gr][ch],pinfo->en[gr][ch+2],
		sizeof(pinfo->en[gr][ch]));
	 memcpy(pinfo->thr_s[gr][ch],pinfo->thr_s[gr][ch+2],
                sizeof(pinfo->thr_s[gr][ch]));
	 memcpy(pinfo->en_s[gr][ch],pinfo->en_s[gr][ch+2],
		sizeof(pinfo->en_s[gr][ch]));
       }
#endif
      }else{
	pe_use[gr][ch]=pe[gr][ch];
      }
    }
  }


  /* bit and noise allocation */
  if (VBR) 
    VBR_iteration_loop( pe_use, ms_ener_ratio, xr, &ratio, &l3_side, l3_enc, 
		    &scalefac, &fr_ps);
  else
    iteration_loop( pe_use, ms_ener_ratio, xr, &ratio, &l3_side, l3_enc, 
		    &scalefac, &fr_ps);

#ifdef BRHIST
  brhist_temp[info->bitrate_index]++;
#endif

  /* flag for our ms_stereo with psy-model on mid & side channels */
  if (force_ms) info->mode_ext = 2;


  /*  write the frame to the bitstream  */
  old_bitrate = info->bitrate_index;
  getframebits(info,stereo,&bitsPerFrame,&mean_bits);
  III_format_bitstream( bitsPerFrame, &fr_ps, l3_enc, &l3_side, 
			&scalefac, &bs, xr, NULL, 0 );


  frameBits = sstell( &bs ) - sentBits;

  
  if ( frameBits % bitsPerSlot )   /* a program failure */
    fprintf( stderr, "Sent %ld bits = %ld slots plus %ld\n",
	     frameBits, frameBits/bitsPerSlot,
	     frameBits%bitsPerSlot );
  sentBits += frameBits;

#ifdef HAVEGTK
  /* copy bit buffer into array and send to mpg123 to synthesis the pcm data */
  if (gtkflag) { 
    mpg123count=0;
    for (i=bs.buf_size-1 ; i > bs.buf_byte_idx ; (i-- ))
      mpg123bs[mpg123count++]=bs.buf[i];
    /* re-synthesize pcm data from mp3 frame */
    if (!decode1(mpg123bs,mpg123count,mpg123pcm,stereo)) {
      if (MPGLAG == MAXMPGLAG) {
	fprintf(stderr,"READ_AHEAD set too low - not enough frame buffering.\n");
	fprintf(stderr,"MP3x display of input and output PCM data out of sync.\n");
      }
      else MPGLAG+=1; 
    }
    /* the mpg123 pcm output stored in this frame will corespond to the
     * pcm input stored in frame computed "mpglag" frames previous.  */
    pinfo->mpglag = MPGLAG;
  }

  
  if (gtkflag) { 
    int adj;
    adj=0;
    if (info->version == 0) adj=576;
    for ( ch = 0; ch < stereo; ch++ ) {
      for ( j = 0; j < WINDELAY; j++ )
	pinfo->pcmdata[ch][j] = pinfo->pcmdata[ch][j+1152-adj];
      for ( j = 0; j < 1152-adj; j++ ) {
	pinfo->pcmdata[ch][j+WINDELAY] = mfbuf[ch][j];
      }
    }
    for ( ch = 0; ch < stereo; ch++ ) {
      for ( j = 0; j < DECDELAY-adj; j++ )
	pinfo->pcmdata2[ch][j] = pinfo->pcmdata2[ch][j+1152-adj];
      for ( j = 0; j < 1152-adj; j++ ) {
	pinfo->pcmdata2[ch][j+DECDELAY-adj] = mpg123pcm[ch][j];
      }
    }
  }
#endif
  /* empty entire bit buffer into mp3 file  */
  empty_buffer(&bs, 1+bs.buf_byte_idx);

  /* to empty buffer w/o file i/o, replace with:
  bs.buf_byte_idx = bs.buf_size -1;
  bs.buf_bit_idx = 8;
  */

  bs.buf[bs.buf_byte_idx] = 0;  

  if (g_bWriteVbrTag) AddVbrFrame(sentBits/8);
  frameNum++;
  return 1152;
}



int lame_readframe(  short int Buffer[2][1152])
{
  int iread = get_audio(Buffer[0],Buffer[1],num_samples,
		      fr_ps.stereo,fr_ps.header);
  return iread;
}







extern void InitFormatBitStream(void);
extern void InitReservoir(void);
/* initialize mp3 encoder */
void lame_init()
{

#ifdef __FreeBSD__
#include <floatingpoint.h>
  {
  /* seet floating point mask to the Linux default */
  fp_except_t mask;
  mask=fpgetmask();
  /*  printf("FreeBSD mask is 0x%x\n",mask); */
  }
#endif
  fprintf(stderr,"LAME version %s (www.sulaco.org/mp3) \n",get_lame_version());
  fprintf(stderr,"GPSYCHO: GPL psycho-acoustic model version %s. \n",get_psy_version());


  frameNum=0;

  // Clear info structure
  memset(&info,0,sizeof(info));
  memset(&bs, 0, sizeof(Bit_stream_struc));
  memset(&fr_ps, 0, sizeof(frame_params));
  memset(&l3_side,0x00,sizeof(III_side_info_t));
  
  fr_ps.header = &info;
  fr_ps.tab_num = -1;             /* no table loaded */
  fr_ps.alloc = NULL;
  info.version = MPEG_AUDIO_ID;   /* =1   Default: MPEG-1 */
  info.extension = 0;       

  InitFormatBitStream();
  InitReservoir();
}



/* clean up if we were producing an mp3 file */
void lame_cleanup()
{
  //int i;

  if (!original_file_type) {
    CloseSndFile();
    III_FlushBitstream();
    /*
    mpg123count=0;
    for (i=bs.buf_size-1 ; i > bs.buf_byte_idx ; (i-- ))
      mpg123bs[mpg123count++]=bs.buf[i];
    */

    empty_buffer(&bs, 1+bs.buf_byte_idx);
    /* to empty buffer w/o file i/o, replace with:
       bs.buf_byte_idx = bs.buf_size -1;
       bs.buf_bit_idx = 8;
    */

    bs.buf[bs.buf_byte_idx] = 0;  

   // Close the file
   fclose(bs.pt);

   // Deallocate all buffers
   desalloc_buffer(&bs);

  }
  
  
  // AF ADDED
  if (g_bWriteVbrTag)
    {
      // Calculate relative quality of VBR stream 
      // 0=best, 100=worst
      int nQuality=VBR_q*100/9;
      
      // Write Xing header again
      PutVbrTag(outPath,nQuality);
    }
  // AF END
}


 
#ifdef BRHIST

#define BRHIST_BARMAX 50

long brhist_count[15];
long brhist_temp[15];
int brhist_vbrmin;
int brhist_vbrmax;
long brhist_max;
char brhist_bps[15][5];
char brhist_backcur[200];
char brhist_bar[BRHIST_BARMAX+10];
char brhist_spc[BRHIST_BARMAX+1];

char stderr_buff[BUFSIZ];


void brhist_init(int br_min, int br_max)
{
  int i;
  char term_buff[1024];
  layer *info = fr_ps.header;
  char *termname;
  char *tp;
  char tc[10];

  for(i = 0; i < 15; i++)
    {
      sprintf(brhist_bps[i], "%3d:", bitrate[info->version][info->lay-1][i]);
      brhist_count[i] = 0;
      brhist_temp[i] = 0;
    }

  brhist_vbrmin = br_min;
  brhist_vbrmax = br_max;

  brhist_max = 0;

  memset(&brhist_bar[0], '*', BRHIST_BARMAX);
  brhist_bar[BRHIST_BARMAX] = '\0';
  memset(&brhist_spc[0], ' ', BRHIST_BARMAX);
  brhist_spc[BRHIST_BARMAX] = '\0';
  brhist_backcur[0] = '\0';

  if ((termname = getenv("TERM")) == NULL)
    {
      fprintf(stderr, "can't get TERM environment string.\n");
      disp_brhist = 0;
      return;
    }

  if (tgetent(term_buff, termname) != 1)
    {
      fprintf(stderr, "can't find termcap entry: %s\n", termname);
      disp_brhist = 0;
      return;
    }

  tc[0] = '\0';
  tp = &tc[0];
  tgetstr("up", &tp);
  brhist_backcur[0] = '\0';
  for(i = br_min-1; i <= br_max; i++)
    strcat(brhist_backcur, tc);

  /*
    tc[0] = '\0';
    tp = &tc[0];
    tgetstr("ce", &tp);
    strcat(brhist_bar, tc);
  */

  setbuf(stderr, stderr_buff);
}

void brhist_add_count(void)
{
  int i;

  for(i = brhist_vbrmin; i <= brhist_vbrmax; i++)
    {
      brhist_count[i] += brhist_temp[i];
      if (brhist_count[i] > brhist_max)
	brhist_max = brhist_count[i];
      brhist_temp[i] = 0;
    }
}

void brhist_disp(void)
{
  int i;
  long full;
  int barlen;

  full = (brhist_max < BRHIST_BARMAX) ? BRHIST_BARMAX : brhist_max;
  fputc('\n', stderr);
  for(i = brhist_vbrmin; i <= brhist_vbrmax; i++)
    {
      barlen = (brhist_count[i]*BRHIST_BARMAX+full-1) / full;
      fputs(brhist_bps[i], stderr);
      fputs(&brhist_bar[BRHIST_BARMAX - barlen], stderr);
      fputs(&brhist_spc[barlen], stderr);
      fputc('\n', stderr);
    }
  fputs(brhist_backcur, stderr);
  fflush(stderr);
}

void brhist_disp_total(void)
{
  int i;
  layer *info = fr_ps.header;

  for(i = brhist_vbrmin; i <= brhist_vbrmax; i++)
    fputc('\n', stderr);

  fprintf(stderr, "----- bitrate statistics -----\n");
  fprintf(stderr, " [kbps]      frames\n");
  for(i = brhist_vbrmin; i <= brhist_vbrmax; i++)
    {
      fprintf(stderr, "   %3d  %8ld (%.1f%%)\n",
	      bitrate[info->version][info->lay-1][i],
	      brhist_count[i],
	      (double)brhist_count[i] / totalframes * 100.0);
    }
  fflush(stderr);
}

#endif


